package	jp.co.sds_corp.ftp;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringReader;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;

public class FtpClient
{
	public enum TransferMode {
		ASCII, BINARY
	}
	Socket	cmdSock;
	OutputStream	os;
	InputStream		is;
	BufferedReader	rd;
	PrintWriter		wr;
	ServerSocket	datSrv = null;
	Socket			datSock = null;
	String			response = null;
	boolean			bDebug = false;
	boolean			bPassive = false;
	InetAddress		datAddr;
	int				nDatPort;
	String	host;
	int	port = 21;
	String	user;
	String	password;
	int	timeout = 30;

	/**
	* コンストラクタ
	*/
	public FtpClient() {
	}

	/**
	* コンストラクタ
	* @param	host		接続先ホスト名
	*/
	public FtpClient( String host )
		throws IOException, FtpServerException
	{
		this.host = host;
	}

	/**
	* コンストラクタ
	* @param	host		接続先ホスト名
	* @param	user		ユーザ名
	* @param	password	パスワード
	*/
	public FtpClient( String host, String user, String password )
		throws IOException, FtpServerException
	{
		this.host = host;
		this.user = user;
		this.password = password;
	}

	/**
	* ホスト(host)を設定します。
	* @param	var		設定するデータ
	*/
	public void setHost( String var ) {
		this.host = var ;
	}
	/**
	* ホスト(host)を取得します。
	* @return	hostの内容
	*/
	public String getHost() {
		return host;
	}

	/**
	* ポート番号(port)を設定します。
	* @param	var		設定するデータ
	*/
	public void setPort( int var ) {
		this.port = var ;
	}
	/**
	* ポート番号(port)を取得します。
	* @return	portの内容
	*/
	public int getPort() {
		return port;
	}

	/**
	* ユーザ名(user)を設定します。
	* @param	var		設定するデータ
	*/
	public void setUser( String var ) {
		this.user = var ;
	}
	/**
	* ユーザ名(user)を取得します。
	* @return	userの内容
	*/
	public String getUser() {
		return user;
	}

	/**
	* パスワード(password)を設定します。
	* @param	var		設定するデータ
	*/
	public void setPassword( String var ) {
		this.password = var ;
	}
	/**
	* パスワード(password)を取得します。
	* @return	passwordの内容
	*/
	public String getPassword() {
		return password;
	}


	/**
	* デバッグモードの設定をします
	* @param	debug		true:デバッグモード false:通常モード
	*/
	public void setDebug( boolean debug ) {
		bDebug = debug;
	}

	/**
	* Passiveモードの設定をします
	* @param	pasv		true:Passiveモード false:通常モード
	*/
	public void setPassive( boolean pasv ) {
		bPassive = pasv;
	}

	/**
	* 送信したコマンドに対する応答メッセージを返します。
	* @return	応答メッセージ
	*/
	public String getResponse() {
		return response;
	}

	/**
	* タイムアウト時間(ミリ秒)(timeout)を設定します。
	* @param	var		設定するデータ
	*/
	public void setTimeout( int var ) {
		this.timeout = var;
	}
	/**
	* タイムアウト時間(ミリ秒)(timeout)を取得します。
	* @return	timeoutの内容
	*/
	public int getTimeout() {
		return timeout;
	}

	/**
	* 送信したコマンドに対する最終のレスポンスコードを返します。
	* @return	レスポンスコード
	*/
	public int getResponseCode() {
		String response = getResponse();
		return parseResponseCode( response );
	}
	/**
	* 応答メッセージからレスポンスコードを解析します
	* @return	レスポンスコード
	*/
	public int parseResponseCode( String response ) {
		int nRes = -1;
		if ( response.length() >= 4 ) {
			try {
				nRes = Integer.valueOf( response.substring( 0, 3 )).intValue();
			} catch ( Exception e ) {
			}
		}
		return nRes;
	}

	/**
	* 送信したコマンドに対するレスポンスを返します。
	* @return	レスポンス
	*/
	public String getMessage() {
		String response = getResponse();
		if ( response.length() >= 4 ) {
			return response.substring( 4 );
		} else {
			return null;
		}
	}

	/**
	* レスポンス文字列からディレクトリ名を切り出す
	*/
	protected String getDirectoryName() {
		String response = getResponse();
		int nStart, nEnd;
		nStart = response.indexOf( '"' );
		nEnd = response.indexOf( '"', nStart + 1 );
		if ( nStart < 0 || nEnd < 0 ) {
			return null;
		}
		return response.substring( nStart + 1, nEnd );
	}

	/**
	* コマンドを送信し、応答を取得します
	*/
	protected void sendCommand( String cmd )
		throws IOException
	{
		writeCommand( cmd );
		readResponse();
	}

	/**
	* コマンドを送信します
	*/
	protected void writeCommand( String cmd )
		throws IOException
	{
		if ( bDebug ) {
			System.out.println( "-->" + cmd );
		}
		// 空読み
		while ( ready()) {
			String buf = rd.readLine();
			if ( buf == null ) {
				break;
			}
			if ( bDebug ) {
				System.out.println( "** " + buf );
			}
		}
		wr.println( cmd );
		wr.flush();
	}

	/**
	* 応答を取得できる状態であるかを確認します
	* @return	true:読み込み可 false:読み込み不可
	*/
	protected boolean ready()
		throws IOException
	{
		try {
			Thread.sleep( 50 );
		} catch ( Exception e ) {
		}
		boolean bRet = is.available() > 0;
		if ( bRet ) {
			return bRet;
		}
		bRet = rd.ready();
		return bRet;
	}

	/**
	* 応答を読み込みます
	*/
	protected void readResponse()
		throws IOException
	{
		response = null;
		StringBuffer res = new StringBuffer();
		String buf;

		int	nWait = 0;
		for ( ;; ) {
			if ( ready() ) {
				break;
			}
			if ( nWait > ( timeout * 10 ) ) {
				throw( new IOException( "timeout" ));
			}
			try {
				Thread.sleep( 100 );
			} catch ( Exception e ) {
			}
			nWait++;
		}

		boolean bMulti = false;
		int nStartRes = -1;
		for ( ;; ) {
			buf = rd.readLine();
			if ( buf == null ) {
				break;
			}
			int nRes = parseResponseCode( buf );
			if ( !bMulti ) {
				if ( nRes >= 0 ) {
					res.append( buf + "\n" );
					if ( buf.charAt( 3 ) == '-' ) {
						bMulti = true;
						nStartRes = nRes;
					} else {
						break;
/*
						if ( !ready() ) {
							break;
						}
*/
					}
				}
			} else {
				res.append( buf + "\n" );
				if ( nRes == nStartRes ) {
					break;
				}
			}
		}
		response = res.toString();
		if ( bDebug ) {
			System.out.print( response );
		}
	}

	/**
	* データ接続の準備を行います
	*/
	protected void prepareDataSocket()
		throws IOException, FtpServerException
	{
		if ( bPassive ) {
			passive();
			datSock = new Socket( datAddr, nDatPort );
		} else {
			port();
		}
	}

	/**
	* データ接続を取得します
	*/
	protected Socket getDataSocket()
		throws IOException, FtpServerException
	{
		Socket sock = null;
		if ( bPassive ) {
			sock = datSock;
		} else {
			sock = datSrv.accept();
		}
		return sock;
	}

	/**
	* データ接続からデータを読み込みます
	*/
	protected void readDataStream( OutputStream out )
		throws IOException, FtpServerException
	{
		Socket sock = getDataSocket();
		InputStream dis = sock.getInputStream();
		byte[] buf = new byte[10240];
		int nRet;
		for ( ;; ) {
			nRet = dis.read( buf, 0, buf.length );
			if ( nRet <= 0 ) {
				break;
			}
			out.write( buf, 0, nRet );
		}
		out.flush();
		dis.close();
		sock.close();
		if ( datSrv != null ) {
			datSrv.close();
		}
		readResponse();
	}

	/**
	* データ接続にデータを書き込みます
	*/
	protected void writeDataStream( InputStream in )
		throws IOException, FtpServerException
	{
		Socket sock = getDataSocket();
		OutputStream out = sock.getOutputStream();
		byte[] buf = new byte[10240];
		int nRet;
		for ( ;; ) {
			nRet = in.read( buf, 0, buf.length );
			if ( nRet <= 0 ) {
				break;
			}
			out.write( buf, 0, nRet );
		}
		out.flush();
		out.close();
		sock.close();
		if ( datSrv != null ) {
			datSrv.close();
		}
		readResponse();
	}

	/**
	* FTP Serverに接続します。
	*/
	public void connect()
		throws IOException, UnknownHostException, FtpServerException
	{
		connect( host, port );
		if ( user != null && password != null ) {
			login( user, password );
		}
	}

	/**
	* 指定されたポートでFTP Serverに接続します。
	* @param	host		接続先ホスト名
	* @param	nPort		接続先ポート番号
	*/
	public void connect( String host, int nPort )
		throws IOException, UnknownHostException, FtpServerException
	{
		cmdSock = new Socket( host, nPort );
		os = cmdSock.getOutputStream();
		wr = new PrintWriter( new BufferedWriter( new OutputStreamWriter(os)));

		is = cmdSock.getInputStream();
		rd = new BufferedReader( new InputStreamReader( is ));

		readResponse();
		switch ( getResponseCode()) {
		case 220: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* FTP Serverとの接続を解除します。
	*/
	public void disconnect()
		throws IOException, UnknownHostException, FtpServerException
	{
		rd.close();
		rd = null;
		wr.close();
		wr = null;
		cmdSock.close();
		cmdSock = null;
	}

	/**
	* FTP Serverにログインします
	* @param	user		ユーザ名
	* @param	password	パスワード
	*/
	public void login( String user, String password )
		throws IOException, FtpServerException
	{
		if ( user != null ) {
			sendCommand( "USER " + user );
			switch ( getResponseCode() ) {
			case 230:
			case 331:
			case 332: break;
			default:
				throw( new FtpServerException( getResponse() ));
			}
			sendCommand( "PASS " + password );
			switch ( getResponseCode() ) {
			case 230:
			case 202:
			case 332: break;
			default:
				throw( new FtpServerException( getResponse() ));
			}
		}
	}

	/**
	* 課金情報を送信します
	* @param	acct		課金情報
	*/
	public void account( String acct )
		throws IOException, FtpServerException
	{
		sendCommand( "ACCT " + acct );
		switch ( getResponseCode() ) {
		case 230:
		case 202: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* データポートを指定します
	*/
	public void port()
		throws IOException, FtpServerException
	{
		datSrv = new ServerSocket( 0, 1 );
		nDatPort = datSrv.getLocalPort();
		if ( bDebug ) {
			System.out.println( "open local port: " + nDatPort );
		}

		InetAddress addr = cmdSock.getLocalAddress();
		int i1 = (int )addr.getAddress()[0];
		int i2 = (int )addr.getAddress()[1];
		int i3 = (int )addr.getAddress()[2];
		int i4 = (int )addr.getAddress()[3];
		if ( i1 < 0 ) {
			i1 += 256;
		}
		if ( i2 < 0 ) {
			i2 += 256;
		}
		if ( i3 < 0 ) {
			i3 += 256;
		}
		if ( i4 < 0 ) {
			i4 += 256;
		}
		sendCommand( "PORT " + i1 + "," + i2 + "," + i3 + "," + i4 + ","
						+ ( nDatPort >> 8 ) + "," + ( nDatPort & 0xff ) );
		switch ( getResponseCode() ) {
		case 200: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* Passiveモードを設定します
	*/
	public void passive()
		throws IOException, FtpServerException
	{
		if ( datSrv != null ) {
			datSrv.close();
			datSrv = null;
		}
		sendCommand( "PASV" );
		switch ( getResponseCode() ) {
		case 227: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
		String res = getMessage();
		int nStart = res.lastIndexOf( '(' ) + 1;
		int nEnd = res.lastIndexOf( ')' );
		String adr = res.substring( nStart, nEnd );
		StringTokenizer tok = new StringTokenizer( adr, "," );
		int nNum = tok.countTokens();
		if ( nNum != 6 ) {
			throw( new FtpServerException( "999 Illegal address format" ));
		}
		String[] buf = new String[nNum];
		int i = 0;
		while ( tok.hasMoreTokens() ) {
			buf[i] = tok.nextToken();
			i++;
		}
		datAddr = InetAddress.getByName(
						buf[0] + "." + buf[1] + "." + buf[2] + "." + buf[3] );
		nDatPort = ( Integer.valueOf( buf[4] ).intValue() << 8 )
					+ Integer.valueOf( buf[5] ).intValue();
		if ( bDebug ) {
			System.out.println( "open remote port: " + datAddr
											+ ":" + nDatPort );
		}
	}

	/**
	* ファイル一覧を取得します
	* @return	ファイル一覧
	*/
	public List<FtpFile> list()
		throws IOException, FtpServerException
	{
		Vector<FtpFile> ret = new Vector<FtpFile>();
		prepareDataSocket();
		sendCommand( "LIST" );
		switch ( getResponseCode() ) {
		case 125:
		case 150: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}

		ByteArrayOutputStream out = new ByteArrayOutputStream();
		readDataStream( out );
		out.close();
		switch ( getResponseCode() ) {
		case 226:
		case 250: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}

		BufferedReader br = new BufferedReader( new StringReader( out.toString()));
		br.readLine();
		for ( ;; ) {
			String line = br.readLine();
			if ( line == null ) {
				break;
			}
			FtpFile file = new FtpFile( line );
			ret.add( file );
		}
		return ret;
	}

	/**
	* ファイルを受信します
	*/
	public void get( String file )
		throws IOException, FtpServerException
	{
		get( file, file );
	}

	/**
	* ファイルを受信します
	*/
	public void get( String remoteFile, String localFile )
		throws IOException, FtpServerException
	{
		FileOutputStream out = new FileOutputStream( localFile );
		get( remoteFile, out );
		out.close();
	}

	/**
	* ファイルを受信します
	*/
	public void get( String remoteFile, OutputStream out )
		throws IOException, FtpServerException
	{
		prepareDataSocket();
		sendCommand( "RETR " + remoteFile );
		switch ( getResponseCode() ) {
		case 125:
		case 150: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}

		readDataStream( out );
		switch ( getResponseCode() ) {
		case 226:
		case 250: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* ファイルを送信します
	*/
	public void put( String file )
		throws IOException, FtpServerException
	{
		put( file, file );
	}

	/**
	* ファイルを送信します
	*/
	public void put( String remoteFile, String localFile )
		throws IOException, FtpServerException
	{
		put( remoteFile, localFile, false );
	}

	/**
	* ファイルを送信します
	*/
	public void put( String remoteFile, String localFile, boolean bUniq )
		throws IOException, FtpServerException
	{
		FileInputStream in = new FileInputStream( localFile );
		put( remoteFile, in, bUniq );
		in.close();
	}

	/**
	* ファイルを送信します
	*/
	public void put( String remoteFile, InputStream in, boolean bUniq )
		throws IOException, FtpServerException
	{
		prepareDataSocket();
		if ( bUniq ) {
			sendCommand( "STOU " + remoteFile );
		} else {
			sendCommand( "STOR " + remoteFile );
		}
		switch ( getResponseCode() ) {
		case 125:
		case 150: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
		writeDataStream( in );
		switch ( getResponseCode() ) {
		case 226:
		case 250: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* ファイルを送信します
	*/
	public void append( String file )
		throws IOException, FtpServerException
	{
		append( file, file );
	}

	/**
	* ファイルを送信します
	*/
	public void append( String remoteFile, String localFile )
		throws IOException, FtpServerException
	{
		FileInputStream in = new FileInputStream( localFile );
		append( remoteFile, in );
		in.close();
	}

	/**
	* ファイルを送信します
	*/
	public void append( String remoteFile, InputStream in )
		throws IOException, FtpServerException
	{
		prepareDataSocket();
		sendCommand( "APPE " + remoteFile );
		switch ( getResponseCode() ) {
		case 125:
		case 150: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}

		writeDataStream( in );
		switch ( getResponseCode() ) {
		case 226:
		case 250: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* ファイル名を変更します
	*/
	public void rename( String from, String to )
		throws IOException, FtpServerException
	{
		sendCommand( "RNFR " + from );
		switch ( getResponseCode() ) {
		case 350: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
		sendCommand( "RNTO " + to );
		switch ( getResponseCode() ) {
		case 250: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* 処理を中断します
	*/
	public void abort()
		throws IOException, FtpServerException
	{
		sendCommand( "ABOR " );
		switch ( getResponseCode() ) {
		case 225:
		case 226: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* ファイルを削除します
	*/
	public void delete( String file )
		throws IOException, FtpServerException
	{
		sendCommand( "DELE " + file );
		switch ( getResponseCode() ) {
		case 250: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* ディレクトリを作成します
	*/
	public void mkdir( String dir )
		throws IOException, FtpServerException
	{
		sendCommand( "MKD " + dir );
		switch ( getResponseCode() ) {
		case 257: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* ディレクトリを削除します
	*/
	public void rmdir( String dir )
		throws IOException, FtpServerException
	{
		sendCommand( "RMD " + dir );
		switch ( getResponseCode() ) {
		case 250: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* SITEパラメータを取得します
	*/
	public String site()
		throws IOException, FtpServerException
	{
		sendCommand( "SITE" );
		switch ( getResponseCode() ) {
		case 200:
		case 202: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
		return getMessage();
	}

	/**
	* 何もしません
	*/
	public void noop()
		throws IOException, FtpServerException
	{
		sendCommand( "NOOP" );
		switch ( getResponseCode() ) {
		case 200: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* システム名称を取得します
	*/
	public String system()
		throws IOException, FtpServerException
	{
		sendCommand( "SYST" );
		switch ( getResponseCode() ) {
		case 215: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
		return getMessage();
	}

	/**
	* ステータスを取得します
	*/
	public String stat()
		throws IOException, FtpServerException
	{
		sendCommand( "STAT" );
		switch ( getResponseCode() ) {
		case 211:
		case 212:
		case 213: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
		return getMessage();
	}

	/**
	* ファイルタイプを設定します
	*/
	public void type( TransferMode mode )
		throws IOException, FtpServerException
	{
		String type;
		switch ( mode ) {
		case ASCII: type = "A"; break;
		case BINARY: type = "I"; break;
		default:
			throw( new IllegalArgumentException());
		}
		sendCommand( "TYPE " + type );
		switch ( getResponseCode() ) {
		case 200: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* ディレクトリを変更します
	*/
	public void cd( String dir )
		throws IOException, FtpServerException
	{
		sendCommand( "CWD " + dir );
		switch ( getResponseCode() ) {
		case 250: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* 一つ上のディレクトリに移動します
	*/
	public void cdup()
		throws IOException, FtpServerException
	{
		sendCommand( "CDUP" );
		switch ( getResponseCode() ) {
		case 200: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
	}

	/**
	* 現在のディレクトリを取得します
	*/
	public String pwd()
		throws IOException, FtpServerException
	{
		sendCommand( "PWD" );
		switch ( getResponseCode() ) {
		case 257: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
		return getDirectoryName();
	}

	/**
	* ヘルプを取得します
	*/
	public String help()
		throws IOException, FtpServerException
	{
		sendCommand( "HELP" );
		switch ( getResponseCode() ) {
		case 211:
		case 214: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
		return getResponse();
	}

	/**
	* 接続をクローズします
	*/
	public void close()
		throws IOException, FtpServerException
	{
		sendCommand( "QUIT" );
		switch ( getResponseCode() ) {
		case 221: break;
		default:
			throw( new FtpServerException( getResponse() ));
		}
		cmdSock.close();
	}

}

